# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Any, TypeAlias, final

from libinephany.observations import observation_utils
from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.pydantic_models.configs.observer_config import ObserverConfig
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates
from libinephany.utils.exceptions import InvalidObservationSizeError
from libinephany.utils.standardizers import Standardizer

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

CachedObservationData: TypeAlias = int | float | list[int | float] | TensorStatistics | dict[str, float | int]

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class Observer(ABC):

    TRAIN = "train"
    INFER = "infer"

    def __init__(
        self,
        *,
        standardizer: Standardizer | None,
        observer_config: ObserverConfig,
        should_standardize: bool = True,
        skip_statistics: list[str] | None = None,
        **kwargs,
    ) -> None:
        """
        :param standardizer: None or the standardizer to apply to the returned observations.
        :param global_config: ObserverConfig that can be used to inform various observation calculations.
        :param should_standardize: Whether standardization should be applied to returned values.
        :param skip_statistics: If the observation uses the TensorStatistic model to return observations, names of the
        fields in the model to not include in returned observations.
        :param kwargs: Miscellaneous keyword arguments.
        """

        self._mode = self.TRAIN
        self._validated_observation = False
        self._cached_observation: CachedObservationData | None = None

        self.observer_config = observer_config
        self.standardize = standardizer if standardizer is not None else observation_utils.null_standardizer
        self.should_standardize = should_standardize and self.can_standardize

        self.skip_statistics = TensorStatistics.filter_skip_statistics(skip_statistics=skip_statistics)

    @final
    @property
    def in_training_mode(self) -> bool:
        """
        :return: Whether the observer is in training mode.
        """

        return self._mode == self.TRAIN

    @final
    @property
    def observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        observation_format = self._get_observation_format()

        if observation_format in StatisticStorageTypes:
            return observation_format

        raise ValueError(
            f"The type of observation format that observations should return must be in the "
            f"{StatisticStorageTypes.__name__} enumeration class. {observation_format} is not and so "
            f"{self.__class__.__name__} is invalid!"
        )

    @property
    def observation_size(self) -> int:
        """
        :return: Number of elements in the observation vector this observation takes up.
        """

        observation_format = self.observation_format

        if observation_format is StatisticStorageTypes.TENSOR_STATISTICS:
            return len([field for field in TensorStatistics.model_fields.keys() if field not in self.skip_statistics])

        elif observation_format is StatisticStorageTypes.FLOAT:
            return 1

        elif observation_format is StatisticStorageTypes.VECTOR:
            return self.vector_length

        else:
            raise ValueError(f"{observation_format} is not a recognised format that observations can return!")

    @property
    def vector_length(self) -> int:
        """
        :return: Length of the vector returned by this observation if it returns a vector.
        """

        raise NotImplementedError

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return True

    @property
    def can_inform(self) -> bool:
        """
        :return: Whether observations from the observer can be used in the agent info dictionary.
        """

        return True

    @property
    @abstractmethod
    def standardizer_key_infix(self) -> str:
        """
        :return: String to infix into the standardizer statistic key to ensure uniqueness.
        """

        raise NotImplementedError

    @abstractmethod
    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        raise NotImplementedError

    @abstractmethod
    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        raise NotImplementedError

    @final
    def _validate_observation(self, observations: list[float | int]) -> None:
        """
        :param observations: Observation vector to validate.
        """

        if len(observations) != self.observation_size:
            raise InvalidObservationSizeError(
                f"The observation vector returned by {self.__class__.__name__} has length {len(observations)} but the "
                f"expected length is {self.observation_size}!"
            )

        self._validated_observation = True

    @abstractmethod
    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        raise NotImplementedError

    @final
    def observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> list[float | int]:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: List of floats or integers to add to the agent's observation vector.
        """

        observations = self._observe(
            observation_inputs=observation_inputs,
            hyperparameter_states=hyperparameter_states,
            tracked_statistics=tracked_statistics,
            action_taken=action_taken,
        )

        if self.can_inform and self._cached_observation is None:
            self._cached_observation = deepcopy(observations)

        if self.observation_format is StatisticStorageTypes.TENSOR_STATISTICS:
            observations = observations.to_list(skip_statistics=self.skip_statistics)  # type: ignore

        observations = [observations] if not isinstance(observations, list) else observations  # type: ignore

        if self.should_standardize:
            observations = [
                self.standardize(
                    statistics_key=f"{self.__class__.__name__}-{self.standardizer_key_infix}-{i}",
                    value_to_standardize=observation,
                )
                for i, observation in enumerate(observations)
            ]

        if not self._validated_observation:
            self._validate_observation(observations=observations)

        return observations

    @final
    def inform(self) -> float | int | dict[str, float] | None:
        """
        :return: The cached observation. If the observation format is TensorStatistics then it is converted to a
        dictionary with the statistics specified in skip_statistics excluded.
        """

        if not self.can_inform:
            return None

        if self._cached_observation is None:
            raise ValueError(
                f"{self.__class__.__name__} cannot inform when no observation has been cached! Ensure observe(...) has "
                f"been called before inform(...)!"
            )

        if self.observation_format is StatisticStorageTypes.TENSOR_STATISTICS:
            observation = self._cached_observation.model_dump(exclude=set(self.skip_statistics))  # type: ignore

        else:
            observation = self._cached_observation

        self._cached_observation = None

        return observation

    @final
    def train(self) -> None:
        """
        Sets the observer to train mode.
        """

        self._mode = self.TRAIN

    @final
    def infer(self) -> None:
        """
        Sets the observer to inference mode.
        """

        self._mode = self.INFER

    def reset(self) -> None:
        """
        Resets the observer.
        """

        pass


class LocalObserver(Observer, ABC):

    def __init__(
        self,
        *,
        agent_id: str,
        parameter_group_name: str | None,
        number_of_discrete_actions: int | None,
        action_scheme_index: int,
        number_of_action_schemes: int,
        **kwargs,
    ) -> None:
        """
        :param agent_id: ID of the agent this observation is assigned to.
        :param parameter_group_name: Name of the parameter group this agent is modulating.
        :param number_of_discrete_actions: The number of discrete actions that can be taken by this agent type's
        policy.
        :param action_scheme_index: Index of the action scheme in use in the list of possible action schemes.
        :param number_of_action_schemes: Number of action schemes available.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(**kwargs)

        self.agent_id = agent_id
        self.parameter_group_name = parameter_group_name

        self.number_of_discrete_actions = number_of_discrete_actions
        self.action_scheme_index = action_scheme_index
        self.number_of_action_schemes = number_of_action_schemes

    @property
    def standardizer_key_infix(self) -> str:
        """
        :return: String to infix into the standardizer statistic key to ensure uniqueness.
        """

        return self.agent_id

    @abstractmethod
    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        raise NotImplementedError

    @abstractmethod
    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        raise NotImplementedError

    @abstractmethod
    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        raise NotImplementedError


class GlobalObserver(Observer, ABC):

    @property
    def standardizer_key_infix(self) -> str:
        """
        :return: String to infix into the standardizer statistic key to ensure uniqueness.
        """

        return "Global"

    @abstractmethod
    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        raise NotImplementedError

    @abstractmethod
    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        raise NotImplementedError

    @abstractmethod
    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        raise NotImplementedError



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import math
from abc import abstractmethod

import numpy as np

from libinephany.utils import random_seeds

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class Sampler:

    def __init__(self, lower_bound: float | int, upper_bound: float | int, **kwargs) -> None:
        """
        :param lower_bound: Lower bound of the distribution to sample from.
        :param upper_bound: Upper bound of the distribution to sample from.
        :param kwargs: Miscellaneous keyword arguments.
        """

        if upper_bound < lower_bound:
            upper_bound, lower_bound = lower_bound, upper_bound

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __call__(
        self,
        number_of_samples: int = 1,
        seed: int | None = None,
    ) -> np.ndarray | float:
        """
        :param number_of_samples: Number of samples to make.
        :param seed: Random seed to use for distribution sampling.
        :return: Array of sampled values or a single sampled value.
        """

        if seed is not None:
            random_seeds.set_all_seeds(seed)

        sample = self.sample(number_of_samples=number_of_samples)

        return sample.item() if sample.size == 1 else sample

    @abstractmethod
    def sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        raise NotImplementedError


class LogUniformSampler(Sampler):

    def sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return sample_from_log_uniform(
            lower_bound=self.lower_bound, upper_bound=self.upper_bound, number_of_samples=number_of_samples
        )


class UniformSampler(Sampler):

    def sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return sample_from_uniform(
            lower_bound=self.lower_bound, upper_bound=self.upper_bound, number_of_samples=number_of_samples
        )


class DiscreteRangeSampler(Sampler):

    def __init__(
        self,
        step: float,
        lower_bound: float | int,
        upper_bound: float | int,
        sample_dtype: type[np.generic | float | int] = np.float64,
        **kwargs,
    ) -> None:
        """
        :param step: Difference between each value in the discrete range.
        :param lower_bound: Lower bound of the distribution to sample from.
        :param upper_bound: Upper bound of the distribution to sample from.
        :param sample_dtype: Datatype the returned samples should have.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(lower_bound=lower_bound, upper_bound=upper_bound, **kwargs)

        self.step = step
        self.sample_dtype = sample_dtype

    def sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return sample_from_discrete_range(
            lower_bound=self.lower_bound,
            upper_bound=self.upper_bound,
            number_of_samples=number_of_samples,
            step=self.step,
        ).astype(self.sample_dtype)


class RoundRobinSampler(Sampler):

    def __init__(self, lower_bound: float | int, upper_bound: float | int, **kwargs) -> None:
        """
        :param lower_bound: Lower bound of the distribution to sample from.
        :param upper_bound: Upper bound of the distribution to sample from.
        :param kwargs: Miscellaneous keyword arguments.
        """
        super().__init__(lower_bound=lower_bound, upper_bound=upper_bound, **kwargs)

        self.magnitude_list = self.generate_magnitude_list()
        self.sampled_elements: set[float] = set()

    def generate_magnitude_list(self) -> list[float]:
        """
        :return: List of possible values that can be sampled in the sampler.
        """

        start_magnitude = math.floor(math.log10(self.upper_bound))
        end_magnitude = math.floor(math.log10(self.lower_bound))

        magnitude_list = []

        for magnitude in range(start_magnitude, end_magnitude - 1, -1):
            for i in range(1, 10):
                value = i * 10**magnitude

                if self.lower_bound <= value <= self.upper_bound:
                    magnitude_list.append(value)

        return magnitude_list

    def sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        if len(self.sampled_elements) == len(self.magnitude_list):
            self.sampled_elements = set()

        sample_choices = [element for element in self.magnitude_list if element not in self.sampled_elements]
        sample = []

        for _ in range(number_of_samples):
            choice = np.random.choice(sample_choices, size=1).item()
            sample_choices.remove(choice)
            sample.append(choice)

        self.sampled_elements.update(sample)

        return np.array(sample)


class DiscreteValueSampler(Sampler):

    def __init__(
        self,
        discrete_values: list[float | int],
        sample_dtype: type[np.generic | float | int] = np.float64,
        **kwargs,
    ) -> None:
        """
        :param discrete_values: List of discrete values to sample from.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(lower_bound=min(discrete_values), upper_bound=max(discrete_values))

        self.discrete_values = discrete_values
        self.sample_dtype = sample_dtype

    def sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        return sample_from_discrete_values(
            discrete_values=self.discrete_values,
            number_of_samples=number_of_samples,
        ).astype(self.sample_dtype)


class RoundRobinDiscreteValueSampler(Sampler):

    def __init__(
        self,
        discrete_values: list[float | int],
        sample_dtype: type[np.generic | float | int] = np.float64,
        **kwargs,
    ) -> None:
        """
        :param discrete_values: List of discrete values to sample from.
        :param kwargs: Miscellaneous keyword arguments.
        """

        super().__init__(lower_bound=min(discrete_values), upper_bound=max(discrete_values))

        self.discrete_values = discrete_values
        self.sample_dtype = sample_dtype

        self.sampled_elements: set[float] = set()

    def sample(self, number_of_samples: int = 1, **kwargs) -> np.ndarray:
        """
        :param number_of_samples: Number of samples to make.
        :param kwargs: Miscellaneous keyword arguments.
        :return: Array of sampled values.
        """

        if len(self.sampled_elements) == len(self.discrete_values):
            self.sampled_elements = set()

        sample_choices = [element for element in self.discrete_values if element not in self.sampled_elements]
        sample = []

        for _ in range(number_of_samples):
            choice = np.random.choice(sample_choices, size=1).item()
            sample_choices.remove(choice)
            sample.append(choice)

        self.sampled_elements.update(sample)

        return np.array(sample)


# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def build_sampler(sampler_name: str, lower_bound: float | int, upper_bound: float | int, **kwargs) -> Sampler:
    """
    :param sampler_name: Name of the sampler to build.
    :param lower_bound: Smallest value that can be sampled by the sampler.
    :param upper_bound: Largest value that can be sampled by the sampler.
    :param kwargs: Miscellaneous keyword arguments that may be required depending on the sampler chosen.
    :return: Constructed sampler.
    """

    possible_samplers = {sampler_type.__name__: sampler_type for sampler_type in Sampler.__subclasses__()}

    try:
        return possible_samplers[sampler_name](lower_bound=lower_bound, upper_bound=upper_bound, **kwargs)  # type: ignore

    except KeyError as e:
        raise ValueError(
            f"Unrecognised sampler {sampler_name}! Possible values: {list(possible_samplers.keys())}."
        ) from e


def sample_from_log_uniform(
    lower_bound: float | int, upper_bound: float | int, number_of_samples: int = 1
) -> np.ndarray:
    """
    :param lower_bound: Lower bound of the distribution.
    :param upper_bound: Upper bound of the distribution.
    :param number_of_samples: Number of samples to make.
    :return: Sampled result.
    """

    sample = np.random.uniform(math.log(lower_bound), math.log(upper_bound), size=number_of_samples)
    sample = np.exp(sample)

    return sample


def sample_from_uniform(lower_bound: float | int, upper_bound: float | int, number_of_samples: int = 1) -> np.ndarray:
    """
    :param lower_bound: Lower bound of the distribution.
    :param upper_bound: Upper bound of the distribution.
    :param number_of_samples: Number of samples to make.
    :return: Sampled result.
    """

    return np.random.uniform(lower_bound, upper_bound, size=number_of_samples)


def sample_from_discrete_range(
    lower_bound: float | int, upper_bound: float | int, step: float, number_of_samples: int = 1
) -> np.ndarray:
    """
    :param lower_bound: Lower bound of the distribution.
    :param upper_bound: Upper bound of the distribution.
    :param step: Difference between each value in the discrete range.
    :param number_of_samples: Number of samples to make.
    :return: Sampled result.
    """

    sample = np.arange(lower_bound, upper_bound + step, step, dtype=np.float64)

    sample = np.random.choice(sample, size=number_of_samples)

    return sample


def sample_from_discrete_values(discrete_values: list[float | int], number_of_samples: int = 1) -> np.ndarray:
    """
    :param discrete_values: List of discrete values to sample from.
    :param number_of_samples: Number of samples to make.
    :return: Sampled result.
    """

    sample = np.random.choice(discrete_values, size=number_of_samples)

    return sample



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from enum import Enum
from itertools import chain
from typing import Any, Callable

import numpy as np
import pandas as pd
import torch
import torch.optim as optim

from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.utils import optim_utils

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

EXP_AVERAGE = "exp_avg"

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class StatisticsCallStage(Enum):

    ON_BATCH_END = "on_batch_end"
    ON_OPTIMIZER_STEP = "on_optimizer_step"
    ON_TRAIN_END = "on_train_end"

    FORWARD_HOOK = "forward_hook"


class StatisticStorageTypes(Enum):

    TENSOR_STATISTICS = TensorStatistics.__name__
    FLOAT = float.__name__
    VECTOR = "vector"


# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def get_exponential_weighted_average(values: list[float]) -> float:
    """
    :param values: List of values to average via EWA.
    :return: EWA of the given values.
    """

    exp_weighted_average = pd.Series(values).ewm(alpha=0.1).mean().iloc[-1]
    assert isinstance(exp_weighted_average, float)

    return exp_weighted_average


def apply_averaging_function_to_tensor_statistics(
    tensor_statistics: list[TensorStatistics], averaging_function: Callable[[list[float]], float]
) -> TensorStatistics:
    """
    :param tensor_statistics: List of statistics models to average over.
    :param averaging_function: Function to average the values with.
    :return: TensorStatistics containing the average over all given tensor statistics.
    """

    fields = TensorStatistics.model_fields.keys()
    averaged_metrics = {
        field: averaging_function([getattr(statistics, field) for statistics in tensor_statistics]) for field in fields
    }

    return TensorStatistics(**averaged_metrics)


def apply_averaging_function_to_dictionary_of_tensor_statistics(
    data: dict[str, list[TensorStatistics]], averaging_function: Callable[[list[float]], float]
) -> dict[str, TensorStatistics]:
    """
    :param data: Dictionary mapping parameter group names to list of TensorStatistics from that parameter group.
    :param averaging_function: Function to average the values with.
    :return: Dictionary mapping parameter group names to TensorStatistics averaged over all statistics in the given
    TensorStatistics models.
    """

    return {
        group: apply_averaging_function_to_tensor_statistics(
            tensor_statistics=metrics, averaging_function=averaging_function
        )
        for group, metrics in data.items()
    }


def apply_averaging_function_to_dictionary_of_metric_lists(
    data: dict[str, list[float]], averaging_function: Callable[[list[float]], float]
) -> dict[str, float]:
    """
    :param data: Dictionary mapping parameter group names to list of metrics from that parameter group.
    :param averaging_function: Function to average the values with.
    :return: Dictionary mapping parameter group names to averages over all metrics from each parameter group.
    """

    return {group: averaging_function(metrics) for group, metrics in data.items()}


def average_tensor_statistics(tensor_statistics: list[TensorStatistics]) -> TensorStatistics:
    """
    :param tensor_statistics: List of TensorStatistics models to average into one model.
    :return: Averages over all given tensor statistics models.
    """

    averaged = {
        field: sum([getattr(statistics_model, field) for statistics_model in tensor_statistics])
        for field in TensorStatistics.model_fields.keys()
    }
    averaged = {field: total / len(tensor_statistics) for field, total in averaged.items()}

    return TensorStatistics(**averaged)


def create_one_hot_observation(vector_length: int, one_hot_index: int | None) -> list[int | float]:
    """
    :param vector_length: Length of the one-hot vector.
    :param one_hot_index: Index of the vector whose element should be set to 1.0, leaving all others as 0.0.
    :return: Constructed one-hot vector in a list.
    """

    if one_hot_index is not None and one_hot_index < 0:
        raise ValueError("One hot indices must be greater than 0.")

    one_hot = np.zeros(vector_length, dtype=np.int8)

    if one_hot_index is not None:
        one_hot[one_hot_index] = 1

    as_list = one_hot.tolist()

    assert isinstance(as_list, list), "One-hot vector must be a list."

    return as_list


def create_one_hot_depth_encoding(agent_controlled_modules: list[str], parameter_group_name: str) -> list[int | float]:
    """
    :param agent_controlled_modules: Ordered list of parameter group names in the inner model.
    :param parameter_group_name: Name of the parameter group to create a depth one-hot vector for.
    :return: Constructed one-hot depth encoding in a list.

    :note: GANNO encodes depths to one-hot vectors of length 3 regardless of the size of the model.
    """

    module_index = agent_controlled_modules.index(parameter_group_name)
    number_of_modules = len(agent_controlled_modules)

    one_hot_index = min(2, (module_index * 3) // number_of_modules)

    return create_one_hot_observation(vector_length=3, one_hot_index=one_hot_index)


def tensor_on_local_rank(tensor: torch.Tensor | None) -> bool:
    """
    :param tensor: Tensor to check whether it is owned by the local rank, partially or entirely.
    :return: Whether the tensor is owned by the local rank.
    """

    return tensor is not None and tensor.grad is not None and tensor.numel() > 0


def form_update_tensor(
    optimizer: optim.Optimizer, parameters: list[torch.Tensor], parameter_group: dict[str, Any]
) -> None | torch.Tensor:
    """
    :param optimizer: Optimizer to form the update tensor from.
    :param parameters: Parameters to create the update tensor from.
    :param parameter_group: Parameter group within the optimizer the given parameters came from.
    :return: None or the formed update tensor.
    """

    if type(optimizer) in optim_utils.ADAM_OPTIMISERS:
        updates_list = [optimizer.state[p][EXP_AVERAGE].view(-1) for p in parameters if tensor_on_local_rank(p)]
        return torch.cat(updates_list) if updates_list else None

    elif type(optimizer) in optim_utils.SGD_OPTIMISERS:
        return optim_utils.compute_sgd_optimizer_update_stats(
            optimizer=optimizer, parameter_group=parameter_group, parameters=parameters
        )

    else:
        raise NotImplementedError(f"Optimizer {type(optimizer).__name__} is not supported!")


def null_standardizer(value_to_standardize: float, **kwargs) -> float:
    """
    :param value_to_standardize: Value to mock the standardization of.
    :return: Given value to standardize.
    """

    return value_to_standardize


def create_sinusoidal_depth_encoding(
    agent_controlled_modules: list[str], parameter_group_name: str, dimensionality: int
) -> list[int | float]:
    """
    :param agent_controlled_modules: Ordered list of parameter group names in the inner model.
    :param parameter_group_name: Name of the parameter group to create a depth encoding for.
    :param dimensionality: Length of the depth vector.
    :return: Sinusoidal depth encoding.
    """

    assert dimensionality % 2 == 0, "Dimensionality of a sinusoidal depth encoding must be even."

    depth = agent_controlled_modules.index(parameter_group_name)

    positions = np.arange(dimensionality // 2)
    frequencies = 1 / (10000 ** (2 * positions / dimensionality))

    encoding = np.zeros(dimensionality)
    encoding[0::2] = np.sin(depth * frequencies)
    encoding[1::2] = np.cos(depth * frequencies)

    vector = encoding.tolist()

    return vector


def concatenate_lists(lists: list[list[Any]]) -> list[Any]:
    """
    :param lists: Lists to concatenate.
    :return: Concatenated lists.
    """

    return list(chain(*lists))
